# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an express
# license agreement from NVIDIA CORPORATION is strictly prohibited.
#

from omni.isaac.kit import SimulationApp

simulation_app = SimulationApp({"headless": False})

from omni.isaac.core import World
from handeye_manipulaotr import HandeyeManipulator
from omni.isaac.manipulators.grippers import ParallelGripper
from omni.isaac.core.utils.stage import add_reference_to_stage
from pick_place_controller import PickPlaceController
from omni.isaac.core.objects import DynamicCuboid
import carb
import sys, os, cv2
import numpy as np
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--test", default=False, action="store_true", help="Run in test mode")
args, unknown = parser.parse_known_args()

my_world = World(stage_units_in_meters=1.0)
my_world.scene.add_default_ground_plane()

working_dir = os.path.dirname(os.path.realpath(__file__))
ur5e_usd_path = os.path.join(working_dir, "ur5e_handeye_gripper_upright.usd")
ur5e_usd_path = os.path.join(working_dir, "ur5e_handeye_gripper.usd")
add_reference_to_stage(usd_path=ur5e_usd_path, prim_path="/World/UR5e")
gripper = ParallelGripper(
    end_effector_prim_path="/World/UR5e/right_inner_finger_pad",
    joint_prim_names=["left_outer_knuckle_joint", "right_outer_knuckle_joint"],
    joint_opened_positions=np.array([0.0, 0.0]),
    joint_closed_positions=np.array([np.pi*2/9, -np.pi*2/9]),
    action_deltas=np.array([-np.pi*2/9, np.pi*2/9]),
)
my_ur5e = my_world.scene.add(
    HandeyeManipulator(
        prim_path="/World/UR5e", name="my_ur5e", end_effector_prim_name="right_inner_finger_pad", gripper=gripper
    )
)
size_scale = 0.03
size_scale_z = 0.03
cube = my_world.scene.add(
    DynamicCuboid(
        name="cube",
        position=np.array([0.4, 0.33, 0.1 + size_scale/2]),
        prim_path="/World/Cube",
        scale=np.array([size_scale, size_scale, size_scale]),
        size=1.0,
        color=np.array([0, 0, 1]),
        mass=0
    )
)
my_world.scene.add_default_ground_plane()
my_ur5e.gripper.set_default_state(my_ur5e.gripper.joint_opened_positions)
my_world.reset()

my_controller = PickPlaceController(
    name="pick_place_controller", gripper=my_ur5e.gripper, robot_articulation=my_ur5e, end_effector_initial_height=0.5
)

rgb_cam = my_ur5e.rgb_cam
rgb_cam.add_instance_segmentation_to_frame()
rgb_cam.initialize()
depth_cam = my_ur5e.depth_cam
depth_cam.add_distance_to_camera_to_frame()
depth_cam.initialize()
image_save_path = f"{os.path.expanduser('~')}/isaac_sim_image"
if not os.path.exists(image_save_path):
    os.makedirs(image_save_path)

articulation_controller = my_ur5e.get_articulation_controller()

i = 0
while simulation_app.is_running():
    my_world.step(render=True)
    if my_world.is_playing():
        if my_world.current_time_step_index == 0:
            my_world.reset()
            my_controller.reset()
        observations = my_world.get_observations()
        actions = my_controller.forward(
            picking_position=cube.get_local_pose()[0],
            placing_position=np.array([0.4, -0.33, 0.02]),
            current_joint_positions=my_ur5e.get_joint_positions(),
            end_effector_offset=np.array([0, 0, 0.25]),
        )

        ''' save image '''
        if i % 100 == 0:
            rgb_image = rgb_cam.get_rgba()[:, :, :3]
            rgb_image = cv2.cvtColor(rgb_image, cv2.COLOR_BGR2RGB)
            depth_image = depth_cam.get_current_frame()["distance_to_camera"]

            rgb_save_path = os.path.join(image_save_path, f"rgb_camera_{i}.png")
            depth_save_path = os.path.join(image_save_path, f"depth_camera_{i}.png")

            try:
                cv2.imwrite(rgb_save_path, rgb_image)
                cv2.imwrite(depth_save_path, depth_image*255)
                # TOOD: depth에 255 255 255가 저장될 때가 있다. 확인 필요
            except:
                print('Pass')
                pass

        if my_controller.is_done():
            print("done picking and placing")
        articulation_controller.apply_action(actions)
    if args.test is True:
        break
    
    i += 1


simulation_app.close()
